import unicodecsv as csv
import json
import sys
import os
import copy
from datetime import datetime

CSV_HEADER_ROW = [
    'SortingIndex',
    'ItemType',
    'Label',
    'Response',
    'Comment',
    'MediaHypertextReference',
    'Latitude',
    'Longitude',
    'ItemScore',
    'ItemMaxScore',
    'ItemScorePercentage',
    'Mandatory',
    'FailedResponse',
    'Inactive',
    'ItemID',
    'ResponseID',
    'ParentID',
    'AuditOwner',
    'AuditAuthor',
    'AuditOwnerID',
    'AuditAuthorID',
    'AuditName',
    'AuditScore',
    'AuditMaxScore',
    'AuditScorePercentage',
    'AuditDuration',
    'DateStarted',
    'DateCompleted',
    'DateModified',
    'AuditID',
    'TemplateID',
    'TemplateName',
    'TemplateAuthor',
    'TemplateAuthorID',
    'ItemCategory',
    'RepeatingSectionParentID',
    'DocumentNo',
    'ConductedOn',
    'PreparedBy',
    'Location',
    'Personnel',
    'ClientSite',
    'AuditSite',
    'AuditArea',
    'AuditRegion',
    'Archived'
]

# audit item empty response 
EMPTY_RESPONSE = ''

# audit item property constants
LABEL = 'label'
COMMENTS = 'comments'
TYPE = 'type'
FAILED = 'failed'
SCORE = 'score'
MAX_SCORE = 'max_score'
SCORE_PERCENTAGE = 'score_percentage'
COMBINED_SCORE = 'combined_score'
COMBINED_MAX_SCORE = 'combined_max_score'
COMBINED_SCORE_PERCENTAGE = 'combined_score_percentage'
PARENT_ID = 'parent_id'
RESPONSE = 'response'
INACTIVE = 'inactive'
ID = 'item_id'
MEDIAID = 'media_id'
HREF = 'href'
EXT = 'file_ext'
SIGNATURE = 'signature'
INFORMATION = 'information'
MEDIA = 'media'
RESPONSES = 'responses'

# maps smartfield conditional statement IDs to the corresponding text
smartfield_conditional_id_to_statement_map = {
    # conditional statements for question field
    '3f206180-e4f6-11e1-aff1-0800200c9a66': 'if response selected',
    '3f206181-e4f6-11e1-aff1-0800200c9a66': 'if response not selected',
    '3f206182-e4f6-11e1-aff1-0800200c9a66': 'if response is',
    '3f206183-e4f6-11e1-aff1-0800200c9a66': 'if response is not',
    '3f206184-e4f6-11e1-aff1-0800200c9a66': 'if response is one of',
    '3f206185-e4f6-11e1-aff1-0800200c9a66': 'if response is not one of',
    # conditional statements for list field
    '35f6c130-e500-11e1-aff1-0800200c9a66': 'if response selected',
    '35f6c131-e500-11e1-aff1-0800200c9a66': 'if response not selected',
    '35f6c132-e500-11e1-aff1-0800200c9a66': 'if response is',
    '35f6c133-e500-11e1-aff1-0800200c9a66': 'if response is not',
    '35f6c134-e500-11e1-aff1-0800200c9a66': 'if response is one of',
    '35f6c135-e500-11e1-aff1-0800200c9a66': 'if response is not one of',
    # conditional statements for slider field
    'cda7c330-e500-11e1-aff1-0800200c9a66': 'if slider value is less than',
    'cda7c331-e500-11e1-aff1-0800200c9a66': 'if slider value is less than or equal to',
    'cda7c332-e500-11e1-aff1-0800200c9a66': 'if slider value is equal to',
    'cda7c333-e500-11e1-aff1-0800200c9a66': 'if slider value is not equal to',
    'cda7c334-e500-11e1-aff1-0800200c9a66': 'if slider value is greater than or equal to',
    'cda7c335-e500-11e1-aff1-0800200c9a66': 'if the slider value is greater than',
    'cda7c336-e500-11e1-aff1-0800200c9a66': 'if the slider value is between',
    'cda7c337-e500-11e1-aff1-0800200c9a66': 'if the slider value is not between',
    # conditional statements for checkbox field
    '4e671f40-e4ff-11e1-aff1-0800200c9a66': 'if the checkbox is checked',
    '4e671f41-e4ff-11e1-aff1-0800200c9a66': 'if the checkbox is not checked',
    # conditional statements for switch field
    '3d346f00-e501-11e1-aff1-0800200c9a66': 'if the switch is on',
    '3d346f01-e501-11e1-aff1-0800200c9a66': 'if the switch is off',
    # conditional statements for text field
    '7c441470-e501-11e1-aff1-0800200c9a66': 'if text is',
    '7c441471-e501-11e1-aff1-0800200c9a66': 'if text is not',
    # conditional statements for textsingle field
    '6ff300f0-e501-11e1-aff1-0800200c9a66': 'if text is',
    '6ff300f1-e501-11e1-aff1-0800200c9a66': 'if text is not',
    # conditional statements for signature field
    '831f8ff0-e500-11e1-aff1-0800200c9a66': 'if signature exists',
    '831f8ff1-e500-11e1-aff1-0800200c9a66': 'if the signature does not exist',
    '831f8ff2-e500-11e1-aff1-0800200c9a66': 'if the signature name is',
    '831f8ff3-e500-11e1-aff1-0800200c9a66': 'if the signature name is not',
    # conditional statements for barcode field
    '8259d900-12e3-11e4-9191-0800200c9a66': 'if the scanned barcode is',
    '8259d901-12e3-11e4-9191-0800200c9a66': 'if the scanned barcode is not'
}

# maps default answer IDs to corresponding Text
standard_response_id_map = {
    '8bcfbf00-e11b-11e1-9b23-0800200c9a66': 'Yes',
    '8bcfbf01-e11b-11e1-9b23-0800200c9a66': 'No',
    '8bcfbf02-e11b-11e1-9b23-0800200c9a66': 'N/A',
    'b5c92350-e11b-11e1-9b23-0800200c9a66': 'Safe',
    'b5c92351-e11b-11e1-9b23-0800200c9a66': 'At Risk',
    'b5c92352-e11b-11e1-9b23-0800200c9a66': 'N/A'
}

# maps header fields to their static IDs
header_field_id = {
    'DocumentNo': 'f3245d46-ea77-11e1-aff1-0800200c9a66',
    'ConductedOn': 'f3245d42-ea77-11e1-aff1-0800200c9a66',
    'PreparedBy': 'f3245d43-ea77-11e1-aff1-0800200c9a66',
    'Location': 'f3245d44-ea77-11e1-aff1-0800200c9a66',
    'Personnel': 'f3245d45-ea77-11e1-aff1-0800200c9a66',
    'ClientSite': 'f3245d41-ea77-11e1-aff1-0800200c9a66'
}


def get_json_property(obj, *args):
    """
    Returns json property if it exists. If it does not exist, returns an empty string
    :param obj:     JON Object
    :param args:    target path, list of keys
    :return:        if path exists, return value, otherwise return empty string
    """
    for arg in args:
        if isinstance(obj, list) and isinstance(arg, int):
            if len(obj) == 0:
                return EMPTY_RESPONSE
            obj = obj[arg]
        elif isinstance(obj, dict) and arg in obj.keys():
            obj = obj[arg]
        else:
            return EMPTY_RESPONSE
    return obj if obj is not None else EMPTY_RESPONSE


class CsvExporter:
    """
    provides tools to convert single json audit to CSV

    Attributes:
        audit_json(json): audit to be converted to CSV
        audit_table(list): the audit data converted to a table
    """

    def __init__(self, audit_json, export_inactive_items=True):
        """
        Constructor

        :param audit_json:      audit in JSON format to be converted to CSV
        """
        self.audit_json = audit_json
        self.export_inactive_items = export_inactive_items
        self.item_category = EMPTY_RESPONSE
        self.repeating_section_title = EMPTY_RESPONSE
        self.item_map = {}
        self.map_items()
        self.audit_table = self.convert_audit_to_table()
        self.all_items = self.audit_items()

    def audit_id(self):
        """
        :return:    The audit ID
        """
        return self.audit_json['audit_id']

    def audit_items(self):
        """
        :return:    All audit items, including header and non-header items
        """
        return self.audit_json['header_items'] + self.audit_json['items']

    def map_items(self):
        """
        Creates a dictionary which maps each item to it's parent ID, Label, and Type.
        This tree can then be traversed recursively to find the Category or Section of a given item.
        """
        for item in self.audit_items():
            if item.get('item_id'):
                self.item_map[item['item_id']] = {
                    'parent_id': item.get('parent_id') or EMPTY_RESPONSE,
                    'item_id': item.get('item_id') or EMPTY_RESPONSE,
                    'label': item.get('label') or EMPTY_RESPONSE,
                    'type': item.get('type') or EMPTY_RESPONSE,
                    'children': item.get('children') or EMPTY_RESPONSE
                }

    def get_item_category(self, item_id):
        """
        Recursively traverses the item Map, following parent IDs until it gets to a Section or Category.
        When a Section or Category is found, the item Category is set to the label of that Section or Category.
        :param item_id: item ID to find Category for
        :return:        Category or Section label 
        """
        if not item_id:
            return EMPTY_RESPONSE
        elif self.item_map[item_id]['type'] == 'section' or self.item_map[item_id]['type'] == 'category' or \
                self.item_map[item_id]['type'] == 'dynamicfield':
            return self.item_map[item_id]['label'] or EMPTY_RESPONSE
        else:
            return self.get_item_category(self.item_map[item_id]['parent_id'])

    def get_repeating_section_title(self, item_id, count):
        """
        Recursively traverses the item Map, following parent IDs until it gets to an element.
        When a Section or Category is found, the item ID is logged - it can then be used to group like questions together.
        Needed whenever questions are nested within a repeating section (e.g. within smart fields.)
        Count used to stop it running longer than needed.
        :param item_id: item ID to find Category for
        :return:        Item ID
        """
        if not item_id or count >= 10:
            return EMPTY_RESPONSE
        elif self.item_map[item_id]['type'] == 'primeelement':
            return 'primeelement'
        elif self.item_map[item_id]['type'] == 'element':
            parent_id = self.item_map[item_id]['parent_id']
            dynamic_field_id = self.item_map[parent_id]['item_id']
            repeating_section_title = self.item_map[parent_id]['label']
            children = self.item_map[parent_id]['children']
            if children != '':
                index = children.index(item_id)
                return dynamic_field_id, repeating_section_title, index
            else:
                return item_id, repeating_section_title, ''

        else:
            count += 1
            return self.get_repeating_section_title(self.item_map[item_id]['parent_id'], count)

    def audit_custom_response_id_to_label_map(self):
        """
        :return:     dictionary mapping custom response_id's to their label
        """
        custom_response_sets = self.audit_json['template_data']['response_sets']
        audit_custom_response_id_to_label_map = dict()
        for response_set in custom_response_sets.keys():
            for response in custom_response_sets[response_set][RESPONSES]:
                audit_custom_response_id_to_label_map[response['id']] = response[LABEL]
        return audit_custom_response_id_to_label_map

    def common_audit_data(self):
        """
        :return:    Selected sub-properties of the audit_data property of the audit JSON as a list
        """
        audit_data_property = self.audit_json['audit_data']
        template_data_property = self.audit_json['template_data']
        audit_date_completed = audit_data_property['date_completed']
        header_data = self.audit_json['header_items']
        audit_data_as_list = list()
        audit_data_as_list.append(audit_data_property['authorship']['owner'])
        audit_data_as_list.append(audit_data_property['authorship']['author'])
        audit_data_as_list.append(audit_data_property['authorship']['owner_id'])
        audit_data_as_list.append(audit_data_property['authorship']['author_id'])
        audit_data_as_list.append(audit_data_property['name'])
        audit_data_as_list.append(audit_data_property[SCORE])
        audit_data_as_list.append(audit_data_property['total_score'])
        audit_data_as_list.append(audit_data_property[SCORE_PERCENTAGE])
        audit_data_as_list.append(audit_data_property['duration'])
        audit_data_as_list.append(self.format_date_time(audit_data_property['date_started']))
        audit_data_as_list.append(self.format_date_time(audit_date_completed))
        audit_data_as_list.append(self.format_date_time(audit_data_property['date_modified']))
        audit_data_as_list.append(self.audit_id())
        audit_data_as_list.append(self.audit_json['template_id'])
        if 'name' in template_data_property[
            'metadata']:  # check for rare case where very old templates/inspections may not include this key
            audit_data_as_list.append(template_data_property['metadata']['name'])
        else:
            audit_data_as_list.append('Untitled Template')
        audit_data_as_list.append(template_data_property['authorship']['author'])
        audit_data_as_list.append(template_data_property['authorship']['author_id'])
        audit_data_as_list.append(self.item_category)
        audit_data_as_list.append(self.repeating_section_title)
        audit_data_as_list.append(self.get_header_item(header_data, 'DocumentNo'))
        audit_data_as_list.append(self.get_header_item(header_data, 'ConductedOn'))
        audit_data_as_list.append(self.get_header_item(header_data, 'PreparedBy'))
        audit_data_as_list.append(self.get_header_item(header_data, 'Location'))
        audit_data_as_list.append(self.get_header_item(header_data, 'Personnel'))
        audit_data_as_list.append(self.get_header_item(header_data, 'ClientSite'))
        audit_data_as_list.append(get_json_property(audit_data_property, 'site', 'name'))
        audit_data_as_list.append(get_json_property(audit_data_property, 'site', 'area', 'name'))
        audit_data_as_list.append(get_json_property(audit_data_property, 'site', 'region', 'name'))
        audit_data_as_list.append(self.audit_json['archived'])
        return audit_data_as_list

    @staticmethod
    def get_header_item(header_data, header_item_type):
        """
        Return standard header item response string
        :param header_data:         Audit JSON array of header items
        :param header_item_type:    Header type whose response is to be returned
        :return:    Header item response string
        """
        for item in header_data:
            if item.get('item_id') == header_field_id.get(header_item_type):
                if 'responses' not in item.keys():
                    return EMPTY_RESPONSE
                if 'text' in item['responses'].keys():
                    return get_json_property(item, 'responses', 'text')
                if 'datetime' in item['responses'].keys():
                    return get_json_property(item, 'responses', 'datetime')
                if 'location_text' in item['responses'].keys():
                    return get_json_property(item, 'responses', 'location_text')
        return EMPTY_RESPONSE

    @staticmethod
    def format_date_time(date):
        """
        Reformat datetime string from ISO format to 'DD Month YYYY HH:MM AM/PM'
        :param date:    date in the format: 2017-03-03T03:45:58.090Z
        :return:        date and time in the format: '03 March 2017 03:45 AM',
        """
        if date:
            date_object = datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.%fZ')
            formatted_date = date_object.strftime('%d %B %Y')
            formatted_time = date_object.strftime('%I:%M:%S %p')
            return formatted_date + ' ' + formatted_time
        else:
            return EMPTY_RESPONSE

    def convert_audit_to_table(self):
        """
        Collects all audit item responses, appends common audit data and returns a 2-dimensional list.
        :return:    2 dimensional list, each list is a single item, which corresponds to a single row
        """
        self.audit_table = []
        for item in self.audit_items():
            if item.get('parent_id'):
                self.item_category = self.get_item_category(item['parent_id'])
                self.repeating_section_title = self.get_repeating_section_title(item['parent_id'], 0)
            else:
                self.item_category = EMPTY_RESPONSE
                self.repeating_section_title = EMPTY_RESPONSE
            row_array = self.item_properties_as_list(item) + self.common_audit_data()
            if get_json_property(item, INACTIVE) and not self.export_inactive_items:
                continue
            self.audit_table.append(row_array)
        return self.audit_table

    def append_converted_audit_to_bulk_export_file(self, output_csv_path):
        """
        Appends audit data table to bulk export file at output_csv_path
        :param output_csv_path: The full path to the file to save
        """
        if not os.path.isfile(output_csv_path) and self.audit_table[0] != CSV_HEADER_ROW:
            self.audit_table.insert(0, CSV_HEADER_ROW)
        self.write_file(output_csv_path, 'ab')

    def save_converted_audit_to_file(self, output_csv_path, allow_overwrite):
        """
        Saves audit data table to a file at output_csv_path
        :param output_csv_path:     The full path to the file to save
        :param allow_overwrite:     if True, allow function to overwrite existing file
        """
        file_exists = os.path.isfile(output_csv_path)
        if file_exists and not allow_overwrite:
            sys.exit(
                'File already exists at ' + output_csv_path +
                '\nPlease set allow_overwrite to True in config.yaml file. See README.md for further instruction')
        elif file_exists and allow_overwrite:
            print('Overwriting file at ' + output_csv_path)
        elif self.audit_table[0] != CSV_HEADER_ROW:
            self.audit_table.insert(0, CSV_HEADER_ROW)
        self.write_file(output_csv_path, 'wb')

    def write_file(self, output_csv_path, mode):
        """
        Saves audit data table to a file at 'path'
        :param output_csv_path: the full path to file to save
        :param mode:    write ('wb') or append ('ab') mode
        """
        try:
            csv_file = open(output_csv_path, mode)
            wr = csv.writer(csv_file, dialect='excel', quoting=csv.QUOTE_ALL)
            wr.writerows(self.audit_table)
            csv_file.close()
        except Exception as ex:
            print(str(ex) + ': Error saving audit_table to ' + output_csv_path)

    def get_item_response(self, item):
        """
        Return item response value
        :param item:    single item in JSON format
        :return:        response property
        """
        response = EMPTY_RESPONSE
        item_type = get_json_property(item, TYPE)
        if item_type == 'question':
            response = get_json_property(item, RESPONSES, 'selected', 0, LABEL)
        elif item_type == 'list':
            for single_response in get_json_property(item, RESPONSES, 'selected'):
                if single_response:
                    response += get_json_property(single_response, LABEL) + '\n'
            response = response[:-1]
        elif item_type == 'address':
            response = get_json_property(item, RESPONSES, 'location_text')
        elif item_type == 'checkbox':
            response = bool(get_json_property(item, RESPONSES, 'value'))
        elif item_type == 'switch':
            response = get_json_property(item, RESPONSES, 'value')
        elif item_type == 'slider':
            response = get_json_property(item, RESPONSES, 'value')
        elif item_type == 'drawing':
            response = get_json_property(item, RESPONSES, 'image', 'media_id')
        elif item_type == MEDIA:
            for image in get_json_property(item, MEDIA):
                response += '\n' + get_json_property(image, 'media_id')
            if response:
                response = response[1:]
        elif item_type == SIGNATURE:
            response = get_json_property(item, RESPONSES, 'name')
        elif item_type == 'smartfield':
            response = get_json_property(item, 'evaluation')
        elif item_type == 'datetime':
            response = self.format_date_time(get_json_property(item, RESPONSES, 'datetime'))
        elif item_type == 'text' or item_type == 'textsingle':
            response = get_json_property(item, RESPONSES, 'text')
        elif item_type == INFORMATION and get_json_property(item, 'options', TYPE) == 'link':
            response = get_json_property(item, 'options', 'link')
        elif item_type == 'temperature':
            response = get_json_property(item, RESPONSES, 'temperature')
        elif item_type in ['dynamicfield', 'element', 'primeelement', 'asset', 'scanner', 'category', 'section',
                           INFORMATION]:
            pass
        else:
            print('Unhandled item type: ' + str(item_type) + ' from ' +
                  self.audit_id() + ', ' + item.get(ID))
        return response

    @staticmethod
    def get_item_response_id(item):
        """
        :param item:    single item in JSON format
        :return:        response ID property
        """
        response_id = EMPTY_RESPONSE
        item_type = get_json_property(item, TYPE)
        if item_type == 'question':
            response_id = get_json_property(item, RESPONSES, 'selected', 0, 'id')
        elif item_type == 'list':
            for single_response in get_json_property(item, RESPONSES, 'selected'):
                if single_response:
                    response_id += get_json_property(single_response, 'id') + '\n'
            response_id = response_id[:-1]
        return response_id

    @staticmethod
    def get_item_score(item):
        """
        retrieve score property from item
        :param item:    single item in JSON format
        :return:        score property or empty string if property does not exist
        """
        if isinstance(get_json_property(item, 'scoring', SCORE), int):
            return get_json_property(item, 'scoring', SCORE)
        elif isinstance(get_json_property(item, 'scoring', COMBINED_SCORE), int):
            return get_json_property(item, 'scoring', COMBINED_SCORE)
        else:
            return EMPTY_RESPONSE

    @staticmethod
    def get_item_max_score(item):
        """
        retrieve max score property from item
        :param item:    single item in JSON format
        :return:        max score property or empty string if property does not exist
        """
        if isinstance(get_json_property(item, 'scoring', MAX_SCORE), int):
            return get_json_property(item, 'scoring', MAX_SCORE)
        elif isinstance(get_json_property(item, 'scoring', COMBINED_MAX_SCORE), int):
            return get_json_property(item, 'scoring', COMBINED_MAX_SCORE)
        else:
            return EMPTY_RESPONSE

    @staticmethod
    def get_item_score_percentage(item):
        """
        retrieve score percentage property from item
        :param item:    single item in JSON format
        :return:        score percentage property or empty string if property does not exist
        """
        if isinstance(get_json_property(item, 'scoring', SCORE_PERCENTAGE), int):
            return get_json_property(item, 'scoring', SCORE_PERCENTAGE)
        elif isinstance(get_json_property(item, 'scoring', COMBINED_SCORE_PERCENTAGE), int):
            return get_json_property(item, 'scoring', COMBINED_SCORE_PERCENTAGE)
        else:
            return EMPTY_RESPONSE

    def get_item_label(self, item):
        """
        retrieve item label property
        :param item:    single item in JSON format
        :return:        label property
        """
        label = EMPTY_RESPONSE
        item_type = get_json_property(item, TYPE)
        if item_type == 'smartfield':
            custom_response_id_to_label_map = self.audit_custom_response_id_to_label_map()
            conditional_id = get_json_property(item, 'options', 'condition')
            if conditional_id:
                label = copy.deepcopy(smartfield_conditional_id_to_statement_map.get(conditional_id)) or EMPTY_RESPONSE
            for value in get_json_property(item, 'options', 'values'):
                label += '|'
                if value in standard_response_id_map.keys():
                    label += standard_response_id_map[value]
                elif value in custom_response_id_to_label_map.keys():
                    label += custom_response_id_to_label_map[value]
                else:
                    label += str(value)
                label += '|'
            return label
        else:
            return get_json_property(item, LABEL)

    @staticmethod
    def get_item_type(item):
        """
        :param item:    single item in JSON format
        :return:        item type property
        """
        item_type = get_json_property(item, TYPE)
        if item_type == INFORMATION:
            item_type += ' - ' + get_json_property(item, 'options', TYPE)
        return item_type

    @staticmethod
    def get_item_media(item):
        """
        :param item:    single item in JSON format
        :return:        item media href links
        """
        item_type = get_json_property(item, TYPE)
        if item_type == INFORMATION and get_json_property(item, 'options', TYPE) == MEDIA:
            media_href = ('{}.{}'.format(get_json_property(item, 'options', MEDIA, MEDIAID),
                                         get_json_property(item, 'options', MEDIA, EXT)))
        elif item_type in ['drawing', SIGNATURE]:
            media_href = ('{}.{}'.format(get_json_property(item, RESPONSES, 'image', MEDIAID),
                                         get_json_property(item, RESPONSES, 'image', EXT)))
        else:
            media_list = []
            for image in get_json_property(item, MEDIA):
                if EXT in image.keys():
                    if image[EXT] is not None:
                        media_list.append(image[MEDIAID] + '.' + image[EXT])
                else:
                    media_list.append(image[MEDIAID] + '.' + 'jpg')
            media_href = '\n'.join(media_list)
            # media_href = '\n'.join(image[MEDIAID]+'.'+image[EXT] for image in get_json_property(item, MEDIA))
        if media_href == '.':
            media_href = None
        return media_href

    @staticmethod
    def get_item_location_coordinates(item):
        """
        :param item:    single item in JSON format
        :return:        comma separated longitude and latitude coordinates
        """
        item_type = get_json_property(item, TYPE)
        if item_type == 'address':
            location_coordinates = get_json_property(item, 'responses', 'location', 'geometry', 'coordinates')
            if isinstance(location_coordinates, list) and len(location_coordinates):
                return str(location_coordinates).strip('[]').split(',')
        return [EMPTY_RESPONSE, EMPTY_RESPONSE]

    def item_properties_as_list(self, item):
        """
        Returns selected properties of the audit item JSON as a list
        :param item:    single item in JSON format
        :return:        array of item data, in format that CSV writer can handle
        """
        location_coordinates = self.get_item_location_coordinates(item)
        latitude = location_coordinates[1]
        longitude = location_coordinates[0]
        return [
            '',
            self.get_item_type(item),
            self.get_item_label(item),
            self.get_item_response(item),
            get_json_property(item, RESPONSES, 'text') if item.get(TYPE) not in ['text',
                                                                                 'textsingle'] else EMPTY_RESPONSE,
            self.get_item_media(item),
            latitude,
            longitude,
            self.get_item_score(item),
            self.get_item_max_score(item),
            self.get_item_score_percentage(item),
            get_json_property(item, 'options', 'is_mandatory') or False,
            get_json_property(item, RESPONSES, FAILED) or False,
            get_json_property(item, INACTIVE) or False,
            get_json_property(item, ID),
            self.get_item_response_id(item),
            get_json_property(item, PARENT_ID)
        ]


def main():
    """
    saves JSON file as CSV. Path to JSON file provided as command line argument
    """
    for arg in sys.argv[1:]:
        audit_json = json.load(open(arg, 'r'))
        csv_exporter = CsvExporter(audit_json)
        csv_exporter.save_converted_audit_to_file(os.path.splitext(arg.split('/')[-1])[0] + '.csv',
                                                  allow_overwrite=True)
    print('Exiting')


if __name__ == '__main__':
    main()



# coding=utf-8
# Author: SafetyCulture
# Copyright: © SafetyCulture 2016
# pylint: disable=E1101

import collections
import json
import logging
import os
import re
import sys
import time
import errno
from builtins import input
from datetime import datetime
import requests
from getpass import getpass

DEFAULT_EXPORT_FORMAT = 'PDF'
GUID_PATTERN = '[A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}$'
HTTP_USER_AGENT_ID = 'safetyculture-python-sdk'


def get_user_api_token(logger):
    """
    Generate iAuditor API Token
    :param logger:  the logger
    :return:        API Token if authenticated else None
    """
    username = input("iAuditor username: ")
    password = getpass()
    generate_token_url = "https://api.safetyculture.io/auth"
    payload = "username=" + username + "&password=" + password + "&grant_type=password"
    headers = {
        'content-type': "application/x-www-form-urlencoded",
        'cache-control': "no-cache",
    }
    response = requests.request("POST", generate_token_url, data=payload, headers=headers)
    if response.status_code == requests.codes.ok:
        return response.json()['access_token']
    else:
        logger.error('An error occurred calling ' + generate_token_url + ': ' + str(response.json()))
        return None


class SafetyCulture:
    def __init__(self, api_token):
        self.current_dir = os.getcwd()
        self.log_dir = self.current_dir + '/log/'
        self.api_url = 'https://api.safetyculture.io/'
        self.audit_url = self.api_url + 'audits/'
        self.template_search_url = self.api_url + 'templates/search?field=template_id&field=name&field=modified_at'
        self.response_set_url = self.api_url + 'response_sets'
        self.get_my_groups_url = self.api_url + 'share/connections'
        self.all_groups_url = self.api_url + 'groups'
        self.add_users_url = self.api_url + 'users'

        self.create_directory_if_not_exists(self.log_dir)
        self.configure_logging()
        logger = logging.getLogger('sp_logger')
        try:
            token_is_valid = re.match('^[a-f0-9]{64}$', api_token)
            if token_is_valid:
                self.api_token = api_token
            else:
                logger.error('API token failed to match expected pattern')
                self.api_token = None
        except Exception as ex:
            self.log_critical_error(ex, 'API token is missing or invalid. Exiting.')
            exit()
        if self.api_token:
            self.custom_http_headers = {
                'User-Agent': HTTP_USER_AGENT_ID,
                'Authorization': 'Bearer ' + self.api_token
            }
        else:
            logger.error('No valid API token parsed! Exiting.')
            sys.exit(1)

    def authenticated_request_get(self, url):
        return requests.get(url, headers=self.custom_http_headers)

    def authenticated_request_post(self, url, data):
        self.custom_http_headers['content-type'] = 'application/json'
        response = requests.post(url, data, headers=self.custom_http_headers)
        del self.custom_http_headers['content-type']
        return response

    def authenticated_request_put(self, url, data):
        self.custom_http_headers['content-type'] = 'application/json'
        response = requests.put(url, data, headers=self.custom_http_headers)
        del self.custom_http_headers['content-type']
        return response

    def authenticated_request_delete(self, url):
        return requests.delete(url, headers=self.custom_http_headers)

    @staticmethod
    def parse_json(json_to_parse):
        """
        Parse JSON string to OrderedDict and return
        :param json_to_parse:  string representation of JSON
        :return:               OrderedDict representation of JSON
        """
        return json.JSONDecoder(object_pairs_hook=collections.OrderedDict).decode(json_to_parse.decode('utf-8'))

    @staticmethod
    def log_critical_error(ex, message):
        """
        Write exception and description message to log

        :param ex:       Exception instance to log
        :param message:  Descriptive message to describe exception
        """
        logger = logging.getLogger('sp_logger')

        if logger is not None:
            logger.critical(message)
            logger.critical(ex)

    def configure_logging(self):
        """
        Configure logging to log to std output as well as to log file
        """
        log_level = logging.WARNING

        log_filename = datetime.now().strftime('%Y-%m-%d') + '.log'
        sp_logger = logging.getLogger('sp_logger')
        sp_logger.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')

        fh = logging.FileHandler(filename=self.log_dir + log_filename)
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        sp_logger.addHandler(fh)

        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(log_level)
        sh.setFormatter(formatter)
        sp_logger.addHandler(sh)

    def create_directory_if_not_exists(self, path):
        """
        Creates 'path' if it does not exist

        If creation fails, an exception will be thrown

        :param path:    the path to ensure it exists
        """
        try:
            os.makedirs(path)
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                self.log_critical_error(ex, 'An error happened trying to create ' + path)
                raise

    def discover_audits(self, template_id=None,
                        modified_after=None,
                        modified_before=None,
                        completed=True,
                        archived=False,
                        limit=1000,
                        order='asc'):
        """
        Return IDs of all completed audits if no parameters are passed, otherwise restrict search
        based on parameter values
        :param archived:        Restrict discover to archived audits
        :param order:           Set whether to return audits in ascending or descending order
        :param limit:           Limit the results returned
        :param template_id:     Restrict discovery to this template_id
        :param modified_after:  Restrict discovery to audits modified after this UTC timestamp
        :param completed:       Restrict discovery to audits marked as completed, default to True
        :return:                JSON object containing IDs of all audits returned by API
        """

        logger = logging.getLogger('sp_logger')
        today = str(datetime.now().replace(microsecond=0).isoformat())
        last_modified = modified_after if modified_after is not None else '2000-01-01T00:00:00.000Z'
        modified_before = modified_before if modified_before is not None else today

        if type(limit) is not int:
            logger.warning('Limit must be an integer. Defaulting to 1000')
            limit = 1000
        else:
            limit = limit
        if order not in ['asc', 'desc']:
            logger.warning('Order can only be ''asc'' or ''desc''. Defaulting to ascending.')
        else:
            order = order
        if type(completed) is not bool:
            logger.warning('Completed must be either true or false. Defaulting to true')
        else:
            completed = completed
        if type(archived) is not bool:
            logger.warning('Archived must be either true or false. Defaulting to false')
        else:
            archived = archived
        search_url = self.audit_url + f'search?field=audit_id&field=modified_at&order={order}&limit={limit}' \
                                      f'&modified_after={last_modified}&modified_before={modified_before}'
        # search_url = self.audit_url + 'search?field=audit_id&field=modified_at&order={}&limit={}&
        # modified_after='.format(order, limit) \
        #     + last_modified
        log_string = '\nInitiating audit_discovery with the parameters: ' + '\n'
        log_string += 'template_id    = ' + str(template_id) + '\n'
        log_string += 'modified_after = ' + str(last_modified) + '\n'
        log_string += 'completed      = ' + str(completed) + '\n'
        logger.info(log_string)

        if template_id is not None:
            log_message = "Searching against {} templates.".format(len(template_id))
            logger.info(log_message)
            if len(template_id) > 1:
                for specific_id in template_id:
                    search_url += '&template=' + specific_id
            else:
                search_url += '&template=' + template_id[0]
                
        if archived is True:
            search_url += '&archived=true'
        if archived is False:
            search_url += '&archived=false'
        if archived == 'both':
            search_url += '&archived=both'

        if completed is True:
            search_url += '&completed=true'
        if completed is False:
            search_url += '&completed=false'
        if completed == 'both':
            search_url += '&completed=both'

        response = self.authenticated_request_get(search_url)
        result = response.json() if response.status_code == requests.codes.ok else None
        number_discovered = str(result['total']) if result is not None else '0'
        log_message = 'on audit_discovery: ' + number_discovered + ' discovered using ' + search_url

        self.log_http_status(response.status_code, log_message)
        return result

    def discover_templates(self, modified_after=None, modified_before=None):
        """
        Query API for all template IDs if no parameters are passed, otherwise restrict search based on parameters

        :param modified_after:   Restrict discovery to templates modified after this UTC timestamp
        :param modified_before:  Restrict discovery to templates modified before this UTC timestamp
        :return:                 JSON object containing IDs of all templates returned by API
        """
        search_url = self.template_search_url
        if modified_before is not None:
            search_url += '&modified_before=' + modified_before
        if modified_after is not None:
            search_url += '&modified_after=' + modified_after

        response = self.authenticated_request_get(search_url)
        result = response.json() if response.status_code == requests.codes.ok else None
        log_message = 'on template discovery using ' + search_url

        self.log_http_status(response.status_code, log_message)
        return result

    def get_preference_ids(self, template_id=None):
        """
        Query API for all preference IDs if no parameters are passed, else restrict to template_id passed
        :param template_id: template_id to obtain export preferences for
        :return:            JSON object containing list of preference objects
        """
        preference_search_url = self.api_url + 'preferences/search'
        if template_id is not None:
            preference_search_url += '?template_id=' + template_id
        response = self.authenticated_request_get(preference_search_url)
        result = response.json() if response.status_code == requests.codes.ok else None
        return result

    def get_export_job_id(self, audit_id, preference_id=None, export_format=DEFAULT_EXPORT_FORMAT):
        """
        Request export job ID from API and return it

        :param audit_id:           audit_id to retrieve export_job_id for
        :param preference_id:      preference to apply to exports
        :param export_format:      desired format of exported document
        :return:                   export job ID obtained from API
        """
        export_url = self.audit_url + audit_id + '/report'
        if export_format == 'docx': # convert old command line format
            export_format = 'WORD'
        export_data = {'format': export_format.upper()}

        if preference_id is not None:
            preference_id_pattern = '^template_[a-fA-F0-9]{32}:' + GUID_PATTERN
            preference_id_is_valid = re.match(preference_id_pattern, preference_id)
            if preference_id_is_valid:
                export_data['preference_id'] = preference_id.split(':')[1]
            else:
                self.log_critical_error(ValueError,
                                        'preference_id {0} does not match expected pattern'.format(
                                            preference_id))

        response = self.authenticated_request_post(export_url, data=json.dumps(export_data))
        result = response.json() if response.status_code == requests.codes.ok else None
        log_message = 'on request to ' + export_url

        self.log_http_status(response.status_code, log_message)
        return result

    def poll_for_export(self, audit_id, export_job_id):
        """
        Poll API for given export job until job is complete or excessive failed attempts occur
        :param audit_id:       audit_id of the export to poll for
        :param export_job_id:  export_job_id of the export to poll for
        :return:               href for export download
        """
        job_id_pattern = '^' + GUID_PATTERN
        job_id_is_valid = re.match(job_id_pattern, export_job_id)

        if job_id_is_valid:
            delay_in_seconds = 5
            poll_url = self.audit_url + audit_id + '/report/' + export_job_id
            export_attempts = 1
            poll_status = self.authenticated_request_get(poll_url)
            status = poll_status.json()
            logger = logging.getLogger('sp_logger')
            if 'status' in status.keys():
                if status['status'] == 'IN_PROGRESS':
                    logger.info(str(status['status']) + ' : ' + audit_id)
                    time.sleep(delay_in_seconds)
                    return self.poll_for_export(audit_id, export_job_id)

                elif status['status'] == 'SUCCESS':
                    logger.info(str(status['status']) + ' : ' + audit_id)
                    return status['url']

                else:
                    if export_attempts < 2:
                        export_attempts += 1
                        logger.info('attempt # {0} exporting report for: ' + audit_id.format(str(export_attempts)))
                        retry_id = self.get_export_job_id(audit_id)
                        return self.poll_for_export(audit_id, retry_id['messageId'])
                    else:
                        logger.error('export for ' + audit_id + ' failed {0} times - skipping'.format(export_attempts))
            else:
                logger.critical('Unexpected response from API: {0}'.format(status))

        else:
            self.log_critical_error(ValueError,
                                    'export_job_id {0} does not match expected pattern'.format(export_job_id))

    def download_export(self, export_href):
        """

        :param export_href:  href for export document to download
        :return:             String representation of exported document
        """

        try:
            response = self.authenticated_request_get(export_href)
            result = response.content if response.status_code == requests.codes.ok else None
            log_message = 'on GET for href: ' + export_href

            self.log_http_status(response.status_code, log_message)
            return result

        except Exception as ex:
            self.log_critical_error(ex, 'Exception occurred while attempting download_export({0})'.format(export_href))

    def get_export(self, audit_id, preference_id=None, export_format=DEFAULT_EXPORT_FORMAT):
        """
        Obtain exported document from API and return string representation of it

        :param audit_id:           audit_id of export to obtain
        :param preference_id:  ID of preference to apply to exports
        :param export_format:      desired format of exported document
        :return:                   String representation of exported document
        """
        export_job_id = self.get_export_job_id(audit_id, preference_id, export_format)['messageId']
        export_href = self.poll_for_export(audit_id, export_job_id)

        export_content = self.download_export(export_href)
        return export_content

    def get_media(self, audit_id, media_id):
        """
        Get media item associated with a specified audit and media ID
        :param audit_id:    audit ID of document that contains media
        :param media_id:    media ID of image to fetch
        :return:            The Content-Type will be the MIME type associated with the media,
                            and the body of the response is the media itself.
        """
        url = self.audit_url + audit_id + '/media/' + media_id
        response = requests.get(url, headers=self.custom_http_headers, stream=True)
        return response

    def get_web_report(self, audit_id):
        """
        Generate Web Report link associated with a specified audit
        :param audit_id:   Audit ID
        :return:           Web Report link
        """
        url = self.audit_url + audit_id + '/web_report_link'
        response = self.authenticated_request_get(url)
        result = self.parse_json(response.content) if response.status_code == requests.codes.ok else None
        self.log_http_status(response.status_code, 'on GET web report for ' + audit_id)
        if result:
            return result.get('url')
        else:
            return None

    def get_audit_actions(self, date_modified, offset=0, page_length=100):
        """
        Get all actions created after a specified date. If the number of actions found is more than 100, this function will
        page until it has collected all actions

        :param date_modified:   ISO formatted date/time string. Only actions created after this date are are returned.
        :param offset:          The index to start retrieving actions from
        :param page_length:     How many actions to fetch for each page of action results
        :return:                Array of action objects
        """
        logger = logging.getLogger('sp_logger')
        actions_url = self.api_url + 'actions/search'
        response = self.authenticated_request_post(
            actions_url,
            data=json.dumps({
                "modified_at": {"from": str(date_modified)},
                "offset": offset,
                "status": [0, 10, 50, 60]
            })
        )
        result = self.parse_json(response.content) if response.status_code == requests.codes.ok else None
        self.log_http_status(response.status_code, 'GET actions')
        if result is None or None in [result.get('count'), result.get('offset'), result.get('total'), result.get('actions')]:
            return None
        return self.get_page_of_actions(logger, date_modified, result, offset, page_length)

    def get_page_of_actions(self, logger, date_modified, previous_page, offset=0, page_length=100):
        """
        Returns a page of action search results

        :param logger: the logger
        :param date_modified: fetch from that date onwards
        :param previous_page: a page of action search results
        :param offset: the index to start retrieving actions from
        :param page_length: the number of actions to return on the search page (max 100)
        :return: Array of action objects
        """
        if previous_page['count'] + previous_page['offset'] < previous_page['total']:
            logger.info('Paging Actions. Offset: ' + str(offset + page_length) + '. Total: ' + str(previous_page['total']))
            next_page = self.get_audit_actions(date_modified, offset + page_length)
            if next_page is None:
                return None
            return next_page + previous_page['actions']
        elif previous_page['count'] + previous_page['offset'] == previous_page['total']:
            return previous_page['actions']

    def get_audit(self, audit_id):
        """
        Request JSON representation of a single specified audit and return it

        :param audit_id:  audit_id of document to fetch
        :return:          JSON audit object
        """
        response = self.authenticated_request_get(self.audit_url + audit_id)
        result = self.parse_json(response.content) if response.status_code == requests.codes.ok else None
        log_message = 'on GET for ' + audit_id

        self.log_http_status(response.status_code, log_message)
        return result

    def create_response_set(self, name, responses):
        """
        Create new response_set
        :param payload:  Name and responses of response_set to create
        :return:
        """
        payload = json.dumps({'name': name, 'responses': responses})
        response = self.authenticated_request_post(self.response_set_url, payload)
        log_message = 'on POST for new response_set: {0}'.format(name)
        self.log_http_status(response.status_code, log_message)

    def get_response_sets(self):
        """
        GET and return all response_sets
        :return: response_sets accessible to user
        """
        response = self.authenticated_request_get(self.response_set_url)
        result = self.parse_json(response.content) if response.status_code == requests.codes.ok else None
        log_message = 'on GET for response_sets'
        self.log_http_status(response.status_code, log_message)
        return result

    def get_response_set(self, responseset_id):
        """
        GET individual response_set by id
        :param responseset_id:  responseset_id of response_set to GET
        :return: response_set
        """
        response = self.authenticated_request_get('{0}/{1}'.format(self.response_set_url, responseset_id))
        result = self.parse_json(response.content) if response.status_code == requests.codes.ok else None
        log_message = 'on GET for {0}'.format(responseset_id)
        self.log_http_status(response.status_code, log_message)
        return result

    def create_response(self, responseset_id, response):
        """
        Create response in existing response_set
        :param responseset_id: id of response_set to add response to
        :param response:       response to add
        :return:               None
        """
        url = '{0}/{1}/responses'.format(self.response_set_url, responseset_id)
        response = self.authenticated_request_post(url, json.dumps(response))
        log_message = 'on POST for new response to: {0}'.format(responseset_id)
        self.log_http_status(response.status_code, log_message)

    def delete_response(self, responseset_id, response_id):
        """
        DELETE individual response by id
        :param responseset_id: responseset_id of response_set containing response to be deleted
        :param response_id:    id of response to be deleted
        :return:               None
        """
        url = '{0}/{1}/responses/{2}'.format(self.response_set_url, responseset_id, response_id)
        response = self.authenticated_request_delete(url)
        log_message = 'on DELETE for response_set: {0}'.format(responseset_id)
        self.log_http_status(response.status_code, log_message)

    def get_my_org(self):
        """
        GET the organisation ID of the requesting user
        :return: The organisation ID of the user
        """
        response = self.authenticated_request_get(self.get_my_groups_url)
        log_message = 'on GET for organisations and groups of requesting user'
        self.log_http_status(response.status_code, log_message)
        my_groups_and_orgs = json.loads(response.content)
        org_id = [group['id'] for group in my_groups_and_orgs['groups'] if group['type'] == "organisation"][0]
        return org_id

    def get_all_groups_in_org(self):
        """
        GET all the groups in the requesting user's organisation
        :return: all the groups of the organisation
        """
        response = self.authenticated_request_get(self.all_groups_url)
        log_message = 'on GET for all groups of organisation'
        self.log_http_status(response.status_code, log_message)
        return response if response.status_code == requests.codes.ok else None

    def get_id_from_email(self, email):
        data = {'email': [email]}
        search_url = self.add_users_url + '/search'
        response = self.authenticated_request_put(search_url, data)
        print(response)
        body = response.content if response.status_code == requests.codes.ok else None
        print(body)
        if body:
            email = body['email']
        else:
            email = None
        print(email)
        return email

    def get_users_of_group(self, group_id):
        """
        GET all the users of the organisations or group
        :param group_id: ID of organisation or group
        :return: array of users
        """
        url = '{0}/{1}/users'.format(self.all_groups_url, group_id)
        response = self.authenticated_request_get(url)
        log_message = 'on GET for users of group: {0}'.format(group_id)
        self.log_http_status(response.status_code, log_message)
        return response.content if response.status_code == requests.codes.ok else None

    def add_user_to_org(self, user_data):
        """
        POST adds a user to organisation
        :param user_data: data of the user to be added
        :return: userID of the user created in the organisation
        """
        url = self.add_users_url
        response = self.authenticated_request_post(url, json.dumps(user_data))
        log_message = 'on POST for adding a user to organisation'
        self.log_http_status(response.status_code, log_message)
        return response.content if response.status_code == requests.codes.ok else None

    def add_user_to_group(self, group_id, user_data):
        """
        POST adds a user to organisation
        :param user_data: contains user ID of user to be added
        :return: userID of the user created in the organisation
        """
        url = '{0}/{1}/users'.format(self.all_groups_url, group_id)
        response = self.authenticated_request_post(url, json.dumps(user_data))
        log_message = 'on POST for adding a user to group'
        self.log_http_status(response.status_code, log_message)
        return response.content if response.status_code == requests.codes.ok else None

    def update_user(self, user_id, user_data):
        """
        PUT updates user details such as user status(active/inactive)
        :param user_id: The ID of the user to update
        :return:  None
        """
        url = '{0}/{1}'.format(self.add_users_url, user_id)
        response = self.authenticated_request_put(url, json.dumps(user_data))
        log_message = 'on PUT for updating a user'
        self.log_http_status(response.status_code, log_message)
        return response if response.status_code == requests.codes.ok else None

    def remove_user(self, role_id, user_id):
        """
        Removes a user from an group/organisation
        :param role_id: The ID of the group or organisation
        :param user_id: The ID of the user to remove
        :return: {ok: true} on successful deletion
        """
        url = '{0}/{1}/users/{2}'.format(self.all_groups_url, role_id, user_id)
        response = self.authenticated_request_delete(url)
        log_message = 'on DELETE for user from group'
        self.log_http_status(response.status_code, log_message)
        return response if response.status_code == requests.codes.ok else None

    @staticmethod
    def log_http_status(status_code, message):
        """
        Write http status code and descriptive message to log

        :param status_code:  http status code to log
        :param message:      to describe where the status code was obtained
        """
        logger = logging.getLogger('sp_logger')
        status_description = requests.status_codes._codes[status_code][0]
        log_string = str(status_code) + ' [' + status_description + '] status received ' + message
        logger.info(log_string) if status_code == requests.codes.ok else logger.error(log_string)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'iart'}

packages = \
['modules']

package_data = \
{'': ['*']}

modules = \
['iart']
install_requires = \
['OrderedDict>=1.1,<2.0',
 'dateparser>=0.7.6,<0.8.0',
 'pandas>=1.0.4,<2.0.0',
 'pyinquirer>=1.0.3,<2.0.0',
 'requests>=2.23.0,<3.0.0',
 'tqdm>=4.46.1,<5.0.0',
 'typer-cli>=0.0.9,<0.0.10',
 'typer>=0.2.1,<0.3.0',
 'unicodecsv>=0.14.1,<0.15.0',
 'xlsxwriter>=1.2.9,<2.0.0']

entry_points = \
{'console_scripts': ['iart_export = iart.iart:export_script',
                     'iart_setup = iart.iart:interactive_setup']}

setup_kwargs = {
    'name': 'iart',
    'version': '0.1.6',
    'description': 'A reporting tool for exporting iAuditor inspections into human readable Excel or CSV files.',
    'long_description': None,
    'author': 'Edd',
    'author_email': 'edward.abrahamsen-mills@safetyculture.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import configparser
import time
from operator import itemgetter

import dateparser
import pandas as pd
import requests
import typer
import xlsxwriter
from PyInquirer import prompt
from dateparser.search import search_dates
from tqdm import tqdm
from xlsxwriter.utility import xl_col_to_name

from modules import safetypy
from modules.RowCreator import RowCreator, TemplateCreator, ConfigSetup
import sys, os
# frozen = 'not'
# if getattr(sys, 'frozen', False):
#         # we are running in a bundle
#         frozen = 'ever so'
#         bundle_dir = sys._MEIPASS
# else:
#         # we are running in a normal Python environment
#         bundle_dir = os.path.dirname(os.path.abspath(__file__))
# print( 'we are',frozen,'frozen')
# print( 'bundle dir is', bundle_dir )
# print( 'sys.argv[0] is', sys.argv[0] )
# print( 'sys.executable is', sys.executable )
# print( 'os.getcwd is', os.getcwd() )
print(sys.argv[0])

if sys.argv[0] not in ['iart_export','iart_setup']:
    os.chdir(os.path.dirname(sys.argv[0]))

app = typer.Typer()

IGNORED_TYPES = ['smartfield', 'primeelement', 'section', 'category', 'element']
COMMON_DATA = ['AuditID', 'Name', 'DateStarted', 'Percentage Score', 'Score', 'Site', 'Area', 'Region',
               'Template Name']
DELIMITER = chr(255)
FAILED_DELIMITER = ' [F]'


def create_excel_document(df, template_id, workbook):
    template_id = template_id[:30]
    worksheet = workbook.add_worksheet(template_id)
    label_col = 0

    template_as_list = df.values.tolist()
    columns_as_list = df.columns.values.tolist()
    response_col = len(columns_as_list) - 1

    # worksheet = workbook.add_worksheet(ft.sanitised_template_name)

    wrap_format = workbook.add_format({'text_wrap': 1})
    header_format = workbook.add_format({'text_wrap': 1, 'rotation': 30})
    failed_item_format = workbook.add_format({'bg_color': '#FFC7CE',
                                              'font_color': '#9C0006'})
    worksheet.set_row(0, 140)

    label_col_letter = xl_col_to_name(label_col)
    response_col_letter = xl_col_to_name(response_col)
    left_border = workbook.add_format()
    left_border.set_left()
    worksheet.set_column(f'{label_col_letter}:{response_col_letter}', 24, wrap_format)

    reformatted_columns = []
    for column in columns_as_list:
        current_row = {'header': column, 'header_format': header_format}
        reformatted_columns.append(current_row)

    table_length = len(template_as_list) + 1
    worksheet.add_table(f'{label_col_letter}1:{response_col_letter}{table_length}', {'name': template_id,
                                                                                     'data': template_as_list,
                                                                                     'columns': reformatted_columns,
                                                                                     'banded_rows': 0,
                                                                                     'banded_columns': 1
                                                                                     })

    worksheet.conditional_format(f'{label_col_letter}1:{response_col_letter}{table_length}',
                                 {'type': 'text',
                                  'criteria': 'containing',
                                  'value': FAILED_DELIMITER,
                                  'format': failed_item_format})


def look_up_col_name(template, column, duplicate_col_check):
    check = template.dict_label_map.get(column)
    if check:
        check_label = check['label'].strip()
        category = check['category'].strip()[:40]
        check_label = " ".join(check_label.split()).strip()
        col_label = (check_label[:200] + '..') if len(check_label) > 200 else check_label
        if category:
            col_label = f'{col_label} ({category})'
        if col_label not in duplicate_col_check:
            duplicate_col_check[col_label] = 1
        else:
            duplicate_col_check[col_label] += 1
            col_label = f"{col_label} ({duplicate_col_check[col_label]})"
        return col_label
    else:
        return column


def rename_cols(df, template, duplicate_col_check):
    for column in tqdm(df.columns):
        rename_to = look_up_col_name(template, column, duplicate_col_check)
        if rename_to != column:
            df.rename(columns={column: rename_to}, inplace=True)

    return df


def convert_to_dict(csv_as_list):
    csv_as_dict = {}

    print('Restructuring Output...')
    for row in tqdm(csv_as_list):
        key = list(row.keys())[0]
        row_values = list(row.values())[0]
        csv_as_dict[key] = row_values

    return csv_as_dict


def process_exports(date_modified, date_before, config):
    sp = config.sp
    list_of_templates = config.templates
    pretty_file_name = time.strftime("%Y-%m-%d at %H%M")
    if list_of_templates:
        if config.file_format == 'Excel':
            workbook = xlsxwriter.Workbook(f'{pretty_file_name}.xlsx')
        for template_id in list_of_templates:
            template_name = [t for t in config.template_lookup if t['value'] == template_id]
            if template_name:
                template_name = template_name[0]['name'].strip()
            else:
                template_name = template_id.strip()
            template_id = template_id.strip()
            styled_template_id = typer.style(template_name, fg=typer.colors.GREEN, bold=True)
            print(f'Beginning processing of {styled_template_id}')
            check_for_inspections = sp.discover_audits(template_id=[template_id], modified_after=date_modified,
                                                       modified_before=date_before, limit=1)
            if check_for_inspections['count'] != 0:
                template = TemplateCreator(template_id, sp, date_modified, date_before)
                inspection_count = len(template.inspection_list['audits'])
                if inspection_count > 0:
                    for inspection in template.full_combined:
                        RowCreator(inspection, sp, template)
                    # for inspection in template.inspection_list['audits']:
                    #     RowCreator(inspection, sp, template)
                    csv_as_dict = convert_to_dict(template.csv_as_list)
                    df = pd.DataFrame(csv_as_dict)
                    print('Renaming Columns...')
                    duplicate_col_check = {}
                    df = rename_cols(df, template, duplicate_col_check)
                    if config.file_format == 'Excel':
                        print('Creating Excel Document...')
                        create_excel_document(df, template_id, workbook)
                    elif config.file_format == 'CSV':
                        df.to_csv(f'{template_name}.csv')
                    print(f'Processing of {styled_template_id} completed.')
                else:
                    print(f'Nothing to export for {styled_template_id}')
                    continue
            else:
                print(f'Nothing to export for {styled_template_id}')
                continue
        if config.file_format == 'Excel':
            workbook.close()
    else:
        typer.echo('No templates specified, you must specify at least one.')
        interactive_setup()


def ask_for_date():
    date_choice = [
        {
            'type': 'list',
            'name': 'date_modified',
            'message': 'Select the date range to export each time the script runs. All times are relative to the '
                       'current date.',
            'choices': [
                '1 Week',
                '2 Weeks',
                '3 Weeks',
                '1 Month',
                'Custom'
            ]
        }
    ]
    answers = prompt(date_choice)

    custom_date_choice = [
        {
            'type': 'input',
            'name': 'custom_date',
            'message': 'What date range to use?'
        }
    ]
    if answers['date_modified'] == 'Custom':
        date_choice = prompt(custom_date_choice)
        date_input = date_choice['custom_date']
        # parsed_date = dateparser.parse(date_input)
        string_searches = date_choice['custom_date']
    else:
        date_input = answers['date_modified']
        # parsed_date = dateparser.parse(date_input)
        string_searches = answers['date_modified']
    parsed_date = date_parser(date_input)
    return parsed_date, string_searches


def set_typer_colour(str_to_format, colour=typer.colors.GREEN, bold=True):
    if type(str_to_format) is str:
        formatted_string = typer.style(str_to_format, fg=colour, bold=bold)
        return formatted_string
    return str_to_format


def date_parser(date):
    parsed_date = dateparser.parse(date)
    if not parsed_date:
        date = search_dates(date)
    else:
        date = parsed_date
    if not date:
        print('The date provided cannot be parsed. Please try again.')
        ask_for_date()
    return date


def get_formatted_list_of_templates(token):
    sp = safetypy.SafetyCulture(token)
    all_templates = sp.discover_templates()
    if all_templates['count'] > 0:
        all_templates = sorted(all_templates['templates'], key=itemgetter('modified_at'), reverse=True)
        template_list = []
        if all_templates:
            for template in all_templates:
                if len(template_list) > 25:
                    break
                else:
                    if template['name'] == '':
                        name = 'Untitled template'
                    else:
                        name = template['name'].strip()
                    template_id = template['template_id']
                    joined_name = f'{name} ({template_id})'
                    template_dict = {'value': template_id, 'name': joined_name, }
                    template_list.append(template_dict)
        else:
            template_list = None
        return template_list
    else:
        return ['No Templates Found.']


def interactive_login():
    login_questions = [
        {
            'type': 'input',
            'name': 'username',
            'message': 'Your iAuditor username (should be your email address.)'
        },
        {
            'type': 'password',
            'name': 'password',
            'message': 'Your iAuditor Password'
        }
    ]
    login = prompt(login_questions)
    username = login['username']
    password = login['password']
    generate_token_url = "https://api.safetyculture.io/auth"
    payload = {
        "username": username,
        "password": password,
        "grant_type": "password"
    }
    headers = {
        'content-type': "application/x-www-form-urlencoded",
        'cache-control': "no-cache",
    }
    response = requests.request("POST", generate_token_url, data=payload, headers=headers)
    if response.status_code == requests.codes.ok:
        typer.echo('Token successfully obtained, continuing to export.')
        return response.json()['access_token']
    else:
        print('An error occurred calling ' + generate_token_url + ': ' + str(response.json()))
        interactive_login()


@app.command()
def interactive_setup():

    iauditor_reporting_tool = set_typer_colour('iAuditor Reporting tool', colour=typer.colors.BRIGHT_MAGENTA,
                                               bold=True)
    api_token_str = set_typer_colour('API token', colour=typer.colors.BRIGHT_MAGENTA, bold=True)
    username_str = set_typer_colour('username', colour=typer.colors.BRIGHT_MAGENTA, bold=True)
    password_str = set_typer_colour('password', colour=typer.colors.BRIGHT_MAGENTA, bold=True)

    welcome = f'Welcome to the {iauditor_reporting_tool}. To get started, we need to generate an {api_token_str}.' \
              f' Please enter your {username_str} and {password_str}:'
    typer.echo(welcome)
    config = ConfigSetup()
    if not config.sp:
        token = interactive_login()
    else:
        overwrite_token = [
        {
            'type': 'confirm',
            'name': 'existing_token',
            'message': 'You already have an API token saved. Do you need to regenerate it?'
        }
        ]
        ask_to_overwrite_token = prompt(overwrite_token)
        if ask_to_overwrite_token['existing_token'] is True:
            token = interactive_login()
        else:
            token = config.token
    templates = get_formatted_list_of_templates(token)
    questions = [
        {
            'type': 'list',
            'name': 'export_format',
            'message': 'Which file format would you like to export?',
            'choices': [
                'Excel',
                'CSV'
            ]
        },
        {
            'type': 'checkbox',
            'name': 'templates',
            'message': 'Which templates to export?',
            'choices': templates
        }
    ]
    answers = prompt(questions)
    date_to_search, string_searches = ask_for_date()
    answers['to_search'] = string_searches
    answers['token'] = token
    answers['templates'] = ','.join(answers['templates'])
    new_config = configparser.ConfigParser()
    new_config['options'] = answers
    with open('config.ini', 'w') as configfile:
        new_config.write(configfile)
    config = ConfigSetup()
    process_exports(str(config.after), str(config.before), config)


@app.command()
def export():
    export_script()


def export_script():
    config = ConfigSetup()
    if config.config:
        typer.echo('Exporting using the same settings as last time.')
        process_exports(str(config.after), str(config.before), config)
    else:
        interactive_setup()


if __name__ == "__main__":
    app()



import configparser
import sys
from collections import OrderedDict
from datetime import datetime, timedelta
from operator import itemgetter

import dateparser
import typer
from dateparser.search import search_dates

from modules import csvExporter, safetypy
from tqdm import tqdm

IGNORED_TYPES = ['smartfield', 'primeelement',
                 'section', 'category',
                 'element', 'information',
                 'information - media', 'information - ']
COMMON_DATA = ['AuditID', 'Name',
               'DateStarted', 'DateCompleted', 'ConductedBy', 'Percentage Score',
               'Score', 'Site', 'Area', 'Region',
               'Template Name']
DELIMITER = chr(255)
values = []
FAILED_DELIMITER = ' [F]'


class TemplateCreator:
    def __init__(self, template_id, sp, date_modified, date_before):
        self.sp = sp
        self.template_id = template_id
        self.most_recent_inspection = self.get_most_recent_inspection()
        self.inspection_json = sp.get_audit(self.most_recent_inspection)
        self.current_table = csvExporter.CsvExporter(self.inspection_json)
        inspection_as_list = self.current_table.convert_audit_to_table()
        self.inspection_as_list = [item for item in inspection_as_list if item[1] not in IGNORED_TYPES]
        self.prime_elements = self.map_prime_elements()
        self.inspection_list = sp.discover_audits(template_id=[template_id], modified_after=date_modified,
                                                  modified_before=date_before)
        self.all_inspections, self.full_combined = self.combine_inspections()
        self.csv_as_list = [
            {'AuditID': []},
            {'Name': []},
            {'DateStarted': []},
            {'DateCompleted': []},
            {'Percentage Score': []},
            {'Score': []},
            {'Site': []},
            {'Area': []},
            {'Region': []},
            {'Template Name': []}
        ]
        self.item_label_map, self.dict_label_map = self.map_item_id_to_label()

    def get_most_recent_inspection(self):
        list_of_inspections = self.sp.discover_audits(template_id=[self.template_id], limit=1, order='desc')
        if list_of_inspections is not None:
            if list_of_inspections['count'] > 0:
                return list_of_inspections['audits'][0]['audit_id']
        else:
            return None

    def combine_inspections(self):
        all_inspections = []
        full_inspections = []
        if self.inspection_list:
            inspection_pbar = tqdm(self.inspection_list['audits'])
            for inspection in inspection_pbar:
                audit_id = inspection['audit_id']
                inspection_pbar.set_description(f"Downloading {audit_id}")
                inspection_json = self.sp.get_audit(audit_id)
                current_table = csvExporter.CsvExporter(inspection_json)
                inspection_as_list = current_table.convert_audit_to_table()
                full_inspections.append(inspection_as_list)
                inspection_as_list = [item for item in inspection_as_list if item[1] not in IGNORED_TYPES]
                all_inspections.append(inspection_as_list)
            return all_inspections, full_inspections
        return None

    def map_prime_elements(self):
        prime_elements = {}
        list_of_prime_elements = [item for item in self.inspection_as_list if item[35] == 'primeelement']
        for pe in list_of_prime_elements:
            prime_elements[pe[14]] = pe[2]
        if prime_elements:
            return prime_elements
        else:
            return None

    def map_item_id_to_label(self):
        checked_dynamicfields = []
        inspection_as_list = self.inspection_as_list
        item_label_map = []
        dict_label_map = OrderedDict()
        count = 0
        for item in inspection_as_list:
            item_id = item[14]
            label = item[2]
            item_type = item[1]
            repeating_section = item[35]
            if item_type == 'dynamicfield':
                if item_id not in checked_dynamicfields:
                    rs_map = self.create_rs_map(item_id)
                    checked_dynamicfields.append(item_id)
                    self.csv_as_list.extend(rs_map)
                    continue
            elif repeating_section in ['']:
                count += 1
                rs_check = False
                new_row = {item_id: []}
                self.csv_as_list.append(new_row)
                item_label_map.append({item_id: {'order': count, 'label': label, 'rs': rs_check}})
                dict_label_map[item_id] = {'order': count, 'label': label, 'rs': rs_check, 'category': item[34]}
            else:
                pass

        return item_label_map, dict_label_map

    def create_rs_map(self, item_id):
        columns = []
        for inspection in self.all_inspections:
            this_inspections_prime_elements = [item for item in inspection if item[35] == 'primeelement'
                                               and item[1] not in IGNORED_TYPES]
            this_inspections_pe_dict = {}
            for pe in this_inspections_prime_elements:
                this_inspections_pe_dict[pe[2]] = pe[14]
            for row in inspection:
                if row[35] not in ['primeelement', '']:
                    parent_id = list(row[35])[0]
                    if parent_id == item_id:
                        # section_title = list(row[35])[1]
                        section_number = list(row[35])[2]
                        category = row[34]
                        category = (category[:50] + '..') if len(category) > 50 else category
                        question = row[2]
                        question = (question[:150] + '..') if len(question) > 150 else question
                        pm_item_id = this_inspections_pe_dict.get(question)
                        new_label = self.prime_elements.get(pm_item_id)
                        if new_label != question:
                            question = new_label
                        new_row_title = f'{category} {section_number} / {question}'
                        new_row = {new_row_title: []}
                        if new_row not in columns:
                            columns.append(new_row)
        return columns


class RowCreator:
    def __init__(self, inspection, sp, template):
        # self.audit_id = inspection['audit_id']
        # self.inspection_json = sp.get_audit(self.audit_id)
        # self.current_table = csvExporter.CsvExporter(self.inspection_json)
        # inspection_as_list = self.current_table.convert_audit_to_table()
        full_combined_inspection = inspection
        self.inspection_as_list = [item for item in full_combined_inspection if item[1] not in IGNORED_TYPES]
        self.list_of_prime_elements = [item for item in full_combined_inspection if item[35] == 'primeelement']
        self.template_prime_elements = template.prime_elements
        self.csv_as_list = template.csv_as_list
        self.mapped_to_list = self.col_to_value()

    def col_to_value(self):
        list_of_prime_elements = self.list_of_prime_elements
        for row in self.csv_as_list:
            response_found = False
            item_id = list(row.keys())[0]
            if item_id in COMMON_DATA:
                item_values = self.inspection_as_list[0]
                common_data_dict = {'AuditID': item_values[29],
                                    'Name': item_values[21],
                                    'DateStarted': item_values[26],
                                    'DateCompleted': item_values[28],
                                    'Percentage Score': item_values[24],
                                    'Score': item_values[22],
                                    'Site': item_values[42],
                                    'Area': item_values[43],
                                    'Region': item_values[44],
                                    'Template Name': item_values[31]
                                    }
                common_response = common_data_dict.get(item_id)
                # print(item_id)
                # print(common_response)
                row[item_id].append(common_response)
                response_found = True
            else:
                search_inspection = [item for item in self.inspection_as_list if item[14] == item_id]
                if search_inspection:
                    response = search_inspection[0][3]
                    failed_status = search_inspection[0][12]
                    if failed_status is True:
                        response += FAILED_DELIMITER
                    row[item_id].append(response)
                    response_found = True
                else:
                    for item in self.inspection_as_list:
                        if item[35] not in ['primeelement', '']:
                            parent_id = list(item[35])[0]
                            section_title = list(item[35])[1]
                            section_number = list(item[35])[2]
                            category = item[34]
                            category = (category[:50] + '..') if len(category) > 50 else category
                            question = item[2]
                            question = (question[:150] + '..') if len(question) > 150 else question
                            response = item[3]
                            for primeelement in list_of_prime_elements:
                                if primeelement[2] == question:
                                    if question != self.template_prime_elements.get(primeelement[14]):
                                        question = self.template_prime_elements.get(primeelement[14])
                            new_row_title = f'{category} {section_number} / {question}'
                            if new_row_title == item_id and response_found is False:
                                row[item_id].append(response)
                                response_found = True
            if response_found is False:
                row[item_id].append('')
        return self.csv_as_list


class ConfigSetup:
    def __init__(self):
        self.config = self.get_config()
        self.sp = self.register_token()
        if self.sp:
            self.token = self.config['options']['token']
            self.templates = self.get_templates()
            self.file_format = self.config['options']['export_format']
            self.to_search = self.config['options']['to_search']
            self.parsed_date = self.date_parser()
            self.before, self.after = self.get_dates_to_search()
            self.template_lookup = self.get_formatted_list_of_templates()

    def get_templates(self):
        if self.config['options']['templates'] is None:
            return []
        if ',' in str(self.config['options']['templates']):
            templates = self.config['options']['templates'].split(',')
        else:
            templates = [self.config['options']['templates']]
        return templates

    @staticmethod
    def get_config():
        config = configparser.ConfigParser()
        config.read('config.ini')
        if config.sections():
            return config
        else:
            return None

    def register_token(self):
        config = self.config
        if self.config:
            if 'options' in config:
                if 'token' in config['options']:
                    sp = safetypy.SafetyCulture(config['options']['token'])
                    return sp
        else:
            return None

    def get_list_of_templates(self):
        list_of_templates = self.sp.discover_templates()
        if list_of_templates is not None:
            list_of_templates = list_of_templates['templates']
            return list_of_templates
        else:
            return None

    def get_formatted_list_of_templates(self):
        all_templates = self.get_list_of_templates()
        all_templates = sorted(all_templates, key=itemgetter('modified_at'), reverse=True)
        template_list = [{'value': 'all', 'name': 'Export All', 'checked': True}]
        if all_templates:
            for template in all_templates:
                if len(template_list) > 20:
                    break
                else:
                    if template['name'] == '':
                        name = 'Untitled template'
                    else:
                        name = template['name'].strip()
                    template_id = template['template_id']
                    joined_name = f'{name} ({template_id})'
                    template_dict = {'value': template_id, 'name': joined_name, }
                    template_list.append(template_dict)
        else:
            template_list = None
        return template_list

    def date_parser(self):
        parsed_date = dateparser.parse(self.to_search)
        if not parsed_date:
            date = search_dates(self.to_search)
        else:
            date = parsed_date
        if not date:
            print('Unable to parse the data in your config file. Defaulting to 1 week ago.')
            date = datetime.now() - timedelta(weeks=1)
        return date

    def get_dates_to_search(self):
        before = datetime.now()
        after = before - timedelta(weeks=1)
        if self.parsed_date:
            if type(self.parsed_date) is list:
                for i, date_to_search in enumerate(self.parsed_date):
                    if i == 0:
                        after = date_to_search[1]
                    if i == 1:
                        before = date_to_search[1]
                    if i > 1:
                        print('More than 2 dates provided, only taking the first two.')
            else:
                after = self.parsed_date

            pretty_after_date = after.strftime(f"%A %d %B %Y at %X")
            pretty_after_date = typer.style(pretty_after_date, fg=typer.colors.GREEN, bold=True)
            pretty_before_date = before.strftime(f"%A %d %B %Y at %X")
            pretty_before_date = typer.style(pretty_before_date, fg=typer.colors.GREEN, bold=True)
            search_from_alert = f'Searching for inspections completed after ' \
                                f'{pretty_after_date} until {pretty_before_date} '
            typer.echo(search_from_alert)
        else:
            print('Either unable to parse date or no date provided. Defaulting to 1 week ago.')
        return before, after
